import win32pdh
import win32process
import win32event
import pywintypes

def timeout_execute (cmd, timeout = 0):
	if timeout == 0:
		timeout = win32event.INFINITE
		
	info  = win32process.CreateProcess(None, cmd, None, None, 0, 0, None, None, win32process.STARTUPINFO())
	subprocess = info [0]
	
	rc = win32event.WaitForSingleObject (subprocess, timeout)			
	
	if rc == win32event.WAIT_FAILED:	
		return -1
		
	if rc == win32event.WAIT_TIMEOUT:
		try:
			win32process.TerminateProcess (subprocess, 0)					
		except pywintypes.error:
			return -3
		return -2
	
	if rc == win32event.WAIT_OBJECT_0:
		return win32process.GetExitCodeProcess(subprocess)


def get_child_pid (cpid):
	object = 'Process'
	items, instances = win32pdh.EnumObjectItems(None, None, object, 
	                                           win32pdh.PERF_DETAIL_WIZARD)
	instance_dict = {}
	for instance in instances:
		try:
			instance_dict[instance] = instance_dict[instance] + 1
		except KeyError:
			instance_dict[instance] = 0
	
	processinfos = []		
	for instance, max_instances in list(instance_dict.items()):
		for inum in range(max_instances+1):
			processinfo = []
			hq = win32pdh.OpenQuery()
			hcs = []
			for item in ['ID Process', 'Creating Process ID']:
				path = win32pdh.MakeCounterPath((None,object,instance,None,inum,item))
				hcs.append(win32pdh.AddCounter(hq,path))
			win32pdh.CollectQueryData(hq)
			processinfo.append (instance[:15].strip ())
			#print "%-15s\t" % (instance[:15]),
			for hc in hcs:
				type,val=win32pdh.GetFormattedCounterValue(hc,win32pdh.PDH_FMT_LONG)
				processinfo.append (int (val))
				#print "%5d" % (val),
				win32pdh.RemoveCounter(hc)
			#print
			win32pdh.CloseQuery(hq)
			processinfos.append (tuple (processinfo))
	
	def recusive (cpid):
		pids = []
		for name, pid, ppid in processinfos:
			if ppid == cpid:
				pids.append (pid)
		for pid in pids:
			pids += recusive (pid)		
		return pids	
	
	pids = recusive (cpid)	
	return pids

if __name__ == "__main__":
	print(get_child_pid (3172))
	




import sys
from google.protobuf import descriptor_pb2

cache = {}

def discover ():
	global cache	
	cache = {}
	for k, v in sys.modules.items ():
		if k.endswith ("_pb2") and not k.startswith ("google."):
			desriptor = v.DESCRIPTOR
			proto = descriptor_pb2.FileDescriptorProto.FromString(desriptor.serialized_pb)
			for service in proto.service:			
				for method in service.method:
					cache ["%s.%s/%s" % (desriptor.package, service.name, method.name)] = (
						v,
						(method.input_type.split (".")[-1], method.client_streaming),
						(method.output_type.split (".")[-1], method.server_streaming),
						method.options
					)					

discover ()

def find_type (uri):
	global cache	
	if not cache: discover ()
	return find_input (uri), find_output (uri)
	
def find_output (uri):
	global cache	
	if not cache: discover ()
	try:
		module, it, ot, opt = cache.get (uri)
	except TypeError:
		raise KeyError	
	return getattr (module, ot [0]), ot [1]

def find_input (uri):
	global cache	
	if not cache: discover ()
	try:
		module, it, ot, opt = cache.get (uri)
	except TypeError:
		raise KeyError	
	return getattr (module, it [0]), it [1]
	



from email.base64mime import body_encode as encode_base64
try:
	from rfc822 import parseaddr
except ImportError:
	from email.utils import parseaddr
import pickle
import time, os
import mimetypes
from hashlib import md5
import shutil

class Composer:
	SAVE_PATH = None
	def __init__ (self, subject, snd, rcpt):		
		self.H = {}
		self.contents = []
		self.H ['MIME-Version'] = '1.0'
		self.H ['Date'] = self.rfc_date ()
		self.H ['X-Priority'] = '3'
		self.H ['X-Mailer'] = 'Skitai SMTP Delivery Agent'
		self.H ['Subject'] = subject.strip ()
		self.H ['From'] = self.snd = snd.strip ()
		self.H ['To'] = self.rcpt = rcpt.strip ()
		self.attachments = []
		self.content_map = {}
		self.smtp_server = None
		self.ssl = False
		self.login = None
		self.saved_name = None
		self.created_time = time.time ()
		self.retrys = 1
	
	def inc_retry (self):
		self.retrys += 1
		
	def set_smtp (self, server, user = None, password = None, ssl = False):
		try: 
			s, p = server.split (":")
			p = int (p)
		except ValueError:
			s, p = server, 25				
		self.smtp_server = (s, p)
		self.ssl = ssl
		if user and password:
			self.login = (user, password)
	
	def set_header (self, name, value):
		self.H [name] = value
		
	def rfc_date (self):
		return time.strftime ('%a, %d %b %Y %H:%M:%S -0000', time.gmtime (time.time ()))
	
	def parse_address (self, addr):
		try:
			return parseaddr(addr)
		except AttributeError:
			return ("", addr)
				
	def encode (self, data):
		return encode_base64 (data)
		
	def add_text (self, data, mimetype, charset = 'utf8'):
		msg = (
			"Content-type: %s; \r\n\tcharset=\"%s\"\r\n"
			"Content-Transfer-Encoding: base64\r\n"
			"\r\n" % (mimetype, charset)
			)
		msg += self.encode (data.encode ("utf8"))
		self.contents.append (msg)
		
	def add_attachment (self, filename, name = None, cid = None):
		if not name:
			name = os.path.split (filename) [-1]			
		if cid:
			self.content_map [name] = cid
		self.attachments.append ((cid, name, filename, mimetypes.guess_type (filename) [0]))
	
	def encode_attachment (self, cid, name, data, mimetype):
		if cid: 
			msg = (
			"Content-ID: <%s>\r\n"
			"Content-Disposition: inline; filename=\"%s\"\r\n"
			"X-Attachment-Id: %s\r\n"
			% (cid, name, cid)
		)
		
		else:
			msg = (
				"Content-Disposition: attachment; filename=\"%s\"\r\n" % name
			)
			
		msg += (
			"MIME-Version: 1.0\r\n"
			"Content-Type: %s; name=\"%s\"\r\n"
			"Content-transfer-encoding: base64\r\n"			
			"\r\n" % (mimetype, name)
			)
		msg += self.encode (data)
		return msg
	
	def remove (self, fn = None):
		if fn is None:
			fn = self.get_FILENAME ()						
		try: os.remove (fn)
		except FileNotFoundError: pass
	
	def moveto (self, path):
		old_path = self.get_FILENAME ()
		self.save (path)
		self.remove (old_path)
			
	def save (self, path):
		while 1:
			d = md5 ((self.get_FROM () + self.get_TO () + str (time.time ())).encode ("utf8"))
			fn = os.path.join (path, "%d.%s" % (self.get_RETRYS (), d.hexdigest ().upper()))
			if not os.path.isfile (fn):
				self.saved_name = fn
				with open (fn, "wb") as f:
					pickle.dump (self, f)				
				break
	
	def send (self):
		self.save (self.SAVE_PATH)
		
	def is_SSL (self):
		return self.ssl
					
	def get_CREATED_TIME (self):
		return self.created_time
				
	def get_TO (self):
		return self.parse_address (self.rcpt) [1]
	
	def get_FROM (self):
		return self.parse_address (self.snd) [1]
	
	def get_SMTP (self):
		return self.smtp_server
	
	def get_LOGIN (self):
		return self.login
	
	def get_RETRYS (self):
		return self.retrys - 1
		
	def get_FILENAME (self):
		return self.saved_name
				
	def get_DATA (self):
		body = "\r\n".join (["%s: %s" % (k, v) for k, v in self.H.items ()]) + "\r\n"		
		
		if len (self.contents) == 0: 
			raise AttributeError		
		
		elif len (self.contents) == 1 and not self.attachments:
			body += self.contents [0]
		
		else:
			sndemail = self.get_FROM ()
			boundary = '__________' + sndemail.replace ('@', '_') + "_" + str (time.time ())
			body += "Content-type: multipart/mixed; \r\n\tboundary=\"%s\"\r\n\r\n" % boundary
			body += "This is a multi-part message in MIME format.\r\n\r\n"
			
			for cid, name, filename, mimetype in self.attachments:
				f = open (filename, "rb")
				data = f.read ()
				f.close ()
				self.contents.append (self.encode_attachment (cid, name, data, mimetype))
				
			for content in self.contents:
				body += "--%s\r\n%s\r\n\r\n" % (boundary, content)
			body += "--%s--\r\n" % boundary
			
		return body


def load (fn):
	with open (fn, "rb") as f:
		m = pickle.load (f)		
	m.inc_retry ()	
	return m

	
if __name__ == "__main__":
	data="""Hi, 
I recieved your message today.

I promise your request is processed with very high priority.

Thanks.
	"""
	m = Composer ("e-Mail Test", '"Tester"<hansroh@lufex.com>', '"Hans Roh"<hansroh@gmail.com>')
	m.set_smtp ("smtp.gmail.com:465", "eunheemax@gmail.com", "!kms2000", True)
	m.add_text (data, "text/html", "utf8")
	m.add_attachment (r"d:/download/setup.py", cid="AAA")
	m.save (r"d:/download")
	
	load (m.get_FILENAME ())
	print (m.get_DATA ())
	print (m.get_FILENAME ())
	m.remove ()
	


import threading, random
from . import asynconnect
import time
try:
	from urllib.parse import urlparse
except ImportError:
	from urlparse import urlparse	
import copy
from aquests.protocols.http2 import H2_PROTOCOLS
from aquests.protocols import http2

PROTO_CONCURRENT_STREAMS = H2_PROTOCOLS

def select_channel (asyncons):
	asyncon = None
	_asyncons = []
	for each in asyncons:
		if each.isactive ():
			continue
		if not each.isconnected ():
			_asyncons = [(each, 0)]
			break
		if each.handler is not None:			
			_asyncons.append ((each, each.handler.jobs ()))
	
	if _asyncons:
		_asyncons.sort (key = lambda x: x [1])
		if _asyncons [0][1] < http2.MAX_HTTP2_CONCURRENT_STREAMS:
			asyncon = _asyncons [0][0]
	
	return asyncon
	

class SocketPool:
	object_timeout = 120
	maintern_interval = 30
	
	def __init__ (self, logger):
		self.__socketfarm = {}
		self.__protos = {}
		self.__numget  = 0
		self.__last_maintern = time.time ()
		self.logger = logger
		self.lock = threading.RLock ()
		self.numobj = 0
	
	def match (self, request):
		return False		
	
	def get_name (self):
		return "__socketpool__"
				
	def status (self):
		info = {}
		cluster = {}
		self.lock.acquire ()
		info ["numget"] = self.__numget
				
		try:
			try:	
				for serverkey, node in list(self.__socketfarm.items ()):	
					nnode = {}
					nnode ["numactives"] = len ([x for x in list(node.values ()) if x.isactive ()])
					nnode ["numconnecteds"] = len ([x for x in list(node.values ()) if x.isconnected ()])
					conns = []
					for asyncon in list(node.values ()):
						stu = {
							"class": asyncon.__class__.__name__, 
							"connected": asyncon.isconnected (), 
							"isactive": asyncon.isactive (), 
							"request_count": asyncon.get_request_count (),
							"event_time": time.asctime (time.localtime (asyncon.event_time)), 
							"zombie_timeout": asyncon.zombie_timeout,								
						}
						try: stu ["has_result"] = asyncon.has_result
						except AttributeError: pass						
						
						if hasattr (asyncon, "get_history"):
							stu ["history"] = asyncon.get_history ()
							
						try:
							stu ["in_map"] = asyncon.is_channel_in_map ()
						except AttributeError: 
							pass	
							
						conns.append (stu)
													
					nnode ["connections"] = conns
					cluster [serverkey] = nnode
					
			finally:
				self.lock.release ()
				
		except:
			self.logger.trace ()
					
		info ["cluster"] = cluster
		return info
		
	def report (self, asyncon, well_functioning):
		pass # for competitable
	
	def get_nodes (self):
		if not self.__socketfarm: return [None] # at least one item needs
		return list(self.__socketfarm.items ())
		
	def maintern (self):		
		# close unused sockets
		for serverkey, node in list(self.__socketfarm.items ()):
			for _id, asyncon in list(node.items ()):					
				if not hasattr (asyncon, "maintern"):
					continue
				
				try:
					deletable = asyncon.maintern (self.object_timeout)					
				except:
					self.logger.trace ()
				else:
					if deletable:
						del self.__socketfarm [serverkey][_id]
						del asyncon
						self.numobj -= 1														
			
			if not self.__socketfarm [serverkey]:
				del self.__socketfarm [serverkey]
				
		self.__last_maintern = time.time ()
				
	def _get (self, serverkey, server, *args):
		asyncon = None
		self.lock.acquire ()
		try:
			try:
				if time.time () - self.__last_maintern > self.maintern_interval:					
					self.maintern ()
							
				self.__numget += 1
				if serverkey not in self.__socketfarm:
					asyncon = self.create_asyncon (server, *args)
					self.__socketfarm [serverkey] = {}
					self.__socketfarm [serverkey][id (asyncon)] = asyncon
					
				else:		
					asyncons = list(self.__socketfarm [serverkey].values ())
					#print ('~~~~~~~~~~', len (asyncons))
					if self.__protos.get (serverkey) in PROTO_CONCURRENT_STREAMS:
						asyncon = select_channel (asyncons)
						
					else:
						for each in asyncons:								
							if not each.isactive ():
								asyncon = each
								break
					
					if not asyncon:
						asyncon = self.create_asyncon (server, *args)
						self.__socketfarm [serverkey][id (asyncon)] = asyncon
				
				asyncon.set_active (True)
			
			finally:
				self.lock.release ()
		
		except:
			self.logger.trace ()
		
		proto = self.__protos.get (serverkey)
		if not proto and asyncon.get_proto ():
			self.__protos [serverkey] = asyncon.get_proto ()
									
		return asyncon
	
	def create_asyncon (self, server, scheme):
		if scheme in ("https", "wss"):
			__conn_class = asynconnect.AsynSSLConnect
			__dft_Port = 443
		elif scheme == "tunnel":
			__conn_class = asynconnect.AsynConnect
			__dft_Port = 443
		elif scheme == "proxy":
			__conn_class = asynconnect.AsynConnect			
			__dft_Port = 5000
		elif scheme == "proxys":
			__conn_class = asynconnect.AsynSSLProxyConnect
			__dft_Port = 5000				
		else:
			__conn_class = asynconnect.AsynConnect
			__dft_Port = 80
		
		try:
			addr, port = server.split (":", 1)
			port = int (port)
		except ValueError:
			addr, port = server, __dft_Port
		
		self.numobj += 1			
		asyncon = __conn_class ((addr, port), self.lock, self.logger)	
		asyncon.set_proxy (scheme == "proxy")
		return asyncon
				
	def get (self, uri):	
		scheme, server, script, params, qs, fragment = urlparse (uri)
		serverkey = "%s://%s" % (scheme, server)	
		return self._get (serverkey, server, scheme)
		
	def cleanup (self):
		self.lock.acquire ()
		try:
			try:
				for server in list(self.__socketfarm.keys ()):					
					for asyncon in list(self.__socketfarm [server].values ()):
						asyncon.disconnect ()
						del asyncon
				self.__socketfarm = {}
				self.__protos = {}
			finally:
				self.lock.release ()
		except:
			self.logger.trace ()


pool = None

def create (logger):
	global pool
	if pool is None:
		pool = SocketPool (logger)

def get (uri):
	return pool.get (uri)
		
def cleanup ():	
	pool.cleanup ()



from . import request_handler
from aquests.protocols.http import tunnel_handler

class ProxyTunnelHandler (request_handler.RequestHandler, tunnel_handler.ProxyTunnelHandler):
	def handle_request (self):	
		if self.asyncon.established:			
			request_handler.RequestHandler.handle_request (self)			
		else:
			tunnel_handler.ProxyTunnelHandler.handle_request (self)
	
	def get_request_buffer (self):		
		if self.asyncon.established:				
			return request_handler.RequestHandler.get_request_buffer (self)			
		else:
			return tunnel_handler.ProxyTunnelHandler.get_request_buffer (self)
		
	def collect_incoming_data (self, data):		
		if self.asyncon.established:
			request_handler.RequestHandler.collect_incoming_data (self, data)	
		else:
			tunnel_handler.ProxyTunnelHandler.collect_incoming_data (self, data)
	
	def found_terminator (self):		
		if self.asyncon.established:
			request_handler.RequestHandler.found_terminator (self)			
		else:
			tunnel_handler.ProxyTunnelHandler.found_terminator (self)			
		
	def found_end_of_body (self):	
		if self.asyncon.established:
			request_handler.RequestHandler.found_end_of_body (self)			
		else:
			tunnel_handler.ProxyTunnelHandler.found_end_of_body (self)
			if self.asyncon.established:
				self._handshaking = True
						

class SSLProxyTunnelHandler (ProxyTunnelHandler, tunnel_handler.ProxyTunnelHandler):
	def convert_to_ssl (self):
		tunnel_handler.ProxyTunnelHandler.convert_to_ssl (self)



# 2016. 1. 10 by Hans Roh hansroh@gmail.com

__version__ = "0.6b1"
version_info = tuple (map (lambda x: not x.isdigit () and x or int (x),  __version__.split (".")))

from . import lifetime, queue, request_builder, response_builder, stubproxy
from .lib import logger as logger_f
from .client import socketpool
from .dbapi import dbpool
from .client import adns
from . import client, dbapi
from .protocols.http import localstorage as ls
from .protocols.http import request_handler
from .protocols import http2
from .dbapi import request as dbo_request
import os
import timeit

try:
	from urllib.parse import urlparse
except ImportError:
	from urlparse import urlparse	


def cb_gateway_demo (response):
	global _logger 
	
	try: cl = len (response.content)
	except: cl = 0	
	if isinstance (response, dbo_request.Request):		
		status = "DBO %s %s %d records/documents received"	% (
			response.code, 
			response.msg,
			cl
		)
	else:
		status = "HTTP/%s %s %s %d bytes received" % (
			response.version,
			response.code, 
			response.msg, 
			cl
		)
			
	_logger.log (
		"REQ %s-%d. %s" % (
		response.meta ['req_method'],
		response.meta ['req_id'],		
		status
		)
	)	
	#print (response.headers)	
	#print (response.data)
	
_request_total = 0			
_finished_total = 0		
_initialized = False
_logger = None
_cb_gateway = cb_gateway_demo
_concurrent = 1
_workers = 1
_currents = 0
_que = queue.Queue ()
_dns_query_req = {}
_timeout = 10

	
def configure (
	workers = 1, logger = None, 
	callback = None, timeout = 10, 
	cookie = False, 
	force_http1 = False,
	http2_constreams = 1,
	allow_redirects = True
):
	global _logger, _cb_gateway, _concurrent, _initialized, _timeout, _workers
	
	if logger is None:
		logger = logger_f.screen_logger ()
	_logger = logger
	
	request_handler.RequestHandler.FORCE_HTTP_11 = force_http1
	request_handler.RequestHandler.ALLOW_REDIRECTS = allow_redirects
	http2.MAX_HTTP2_CONCURRENT_STREAMS = max (http2_constreams, 3)		
	_workers = workers
	_concurrent = workers
	if not force_http1:
		_concurrent = workers * http2_constreams
			
	if callback:
		_cb_gateway = callback
	
	if cookie:
		ls.create (_logger)
	_timeout = timeout	
	client.set_timeout (timeout)
	dbapi.set_timeout (timeout)
	
	socketpool.create (_logger)
	dbpool.create (_logger)
	adns.init (_logger)
	lifetime.init (_timeout / 2.) # maintern interval
	_initialized = True
	

def _next ():
	global _currents, _concurrent, _finished_total
	
	_finished_total += 1
	if lifetime._shutdown_phase:
		return
				
	#print ('===========', _concurrent, _currents)
	for i in range (_concurrent - _currents):
		_req ()
			
def _request_finished (handler):
	global _cb_gateway, _currents, _concurrent, _finished_total, _logger
	
	_currents -= 1
	if isinstance (handler, dbo_request.Request):
		response = handler
	else:
		response = response_builder.HTTPResponse (handler)
		
	response.logger = _logger
	_cb_gateway (response)
	_next ()

def _add (method, url, params = None, auth = None, headers = {}, meta = {}, proxy = None):
	def dns_result (answer = None):
		pass
		
	global _que, _initialized, _dns_query_req			
	
	if not _initialized:		
		configure ()		
	if not meta: meta = {}
	meta ['req_id'] = _que.req_id
	meta ['req_method'] = method
	host = urlparse (url) [1].split (":")[0]
	# DNS query for caching and massive 
	if host not in _dns_query_req:
		_dns_query_req [host] = None
		adns.query (host, "A", callback = dns_result)	
	_que.add ((method, url, params, auth, headers, meta, proxy))


def _req ():
	global _que, _logger, _finished_total, _currents, _request_total
	
	args = _que.get ()
	if args is None:
		if lifetime._shutdown_phase == 0:		
			lifetime.shutdown (0, 7)
		return
	
	_currents += 1
	_request_total += 1
	_method = args [0].lower ()
	if _method in ("postgresql", "redis", "mongodb", "sqlite3"):
		method, server, (dbmethod, params), dbname, auth, meta = args
		asyncon = dbpool.get (server, dbname, auth, "*" + _method)
		req = request_builder.make_dbo (_method, server, dbmethod, params, dbname, meta, _logger)
		req.set_callback (_request_finished)
		asyncon.execute (req)
		
	else:	
		method, url, params, auth, headers, meta, proxy = args
		asyncon = socketpool.get (url)		
		if _method == "ws":
			req, handler_class = request_builder.make_ws (_method, url, params, auth, headers, meta, proxy, _logger)		
			handler = handler_class (asyncon, req, _request_finished)
			
		else:
			req, handler_class = request_builder.make_http (_method, url, params, auth, headers, meta, proxy, _logger)
			handler = handler_class (asyncon, req, _request_finished)
		if asyncon.get_proto () and asyncon.isconnected ():			
			asyncon.handler.handle_request (handler)
		else:
			handler.handle_request ()
		
def countreq ():
	global _request_total
	return 	_request_total

def countfin ():	
	global _finished_total
	return _finished_total

def countcli ():	
	global _currents
	return _currents

def fetchall ():
	global _workers, _logger, _que, _timeout
	
	if not _initialized:
		configure ()
	
	_fetch_started = timeit.default_timer ()
	for i in range (_workers):
		_req ()		
		
	lifetime.loop (os.name == "nt" and 1.0 or _timeout / 2.0)
	_duration = timeit.default_timer () - _fetch_started	
	socketpool.cleanup ()
	dbpool.cleanup ()
	
	_logger.log ("* %d tasks during %1.5f sec, %1.2f tasks/sec" % (_que.req_id, _duration, _que.req_id / _duration))
	
	
#----------------------------------------------------
# REST CALL
#----------------------------------------------------	
def get (*args, **karg):
	_add ('get', *args, **karg)

def delete (*args, **karg):
	_add ('delete', *args, **karg)

def head (*args, **karg):
	_add ('head', *args, **karg)

def trace (*args, **karg):
	_add ('trace', *args, **karg)

def options (*args, **karg):
	_add ('options', *args, **karg)
				
def post (*args, **karg):
	_add ('post', *args, **karg)

def put (*args, **karg):
	_add ('put', *args, **karg)
			
def postform (*args, **karg):
	_add ('postform', *args, **karg)

def postxml (*args, **karg):
	_add ('postxml', *args, **karg)

def postjson (*args, **karg):
	_add ('postjson', *args, **karg)	

def postnvp (*args, **karg):
	_add ('postnvp', *args, **karg)	

def putform (*args, **karg):
	_add ('putform', *args, **karg)

def putxml (*args, **karg):
	_add ('putxml', *args, **karg)

def putjson (*args, **karg):
	_add ('putjson', *args, **karg)	

def putnvp (*args, **karg):
	_add ('putnvp', *args, **karg)	
	
def upload (*args, **karg):
	_add ('upload', *args, **karg)	

def ws (*args, **karg):
	_add ('ws', *args, **karg)	

#----------------------------------------------------
# XMLRPC, gRPC
#----------------------------------------------------
def _addrpc (method, rpcmethod, params, url, auth = None, headers = {}, meta = {}, proxy = None):	
	_add (method, url, (rpcmethod, params), auth, headers, meta, proxy)
	
def rpc	(*args, **karg):
	return stubproxy.Proxy ('rpc', _addrpc, *args, **karg)
	
def grpc	(*args, **karg):
	return stubproxy.Proxy ('grpc', _addrpc, *args, **karg)

#----------------------------------------------------
# DBO QEURY
#----------------------------------------------------
def _adddbo (method, dbmethod, params, server, dbname = None, auth = None, meta = {}):
	global _que
	
	if not _initialized:
		configure ()
	if not meta: meta = {}	
	meta ['req_id'] = _que.req_id
	meta ['req_method'] = method
	_que.add ((method, server, (dbmethod, params), dbname, auth, meta))
	
def postgresql (*args, **karg):
	return stubproxy.Proxy ('postgresql', _adddbo, *args, **karg)
pg = postgresql

def redis (*args, **karg):
	return stubproxy.Proxy ('redis', _adddbo, *args, **karg)

def mongodb (*args, **karg):
	return stubproxy.Proxy ('mongodb', _adddbo, *args, **karg)
	
def sqlite3 (*args, **karg):
	return stubproxy.Proxy ('sqlite3', _adddbo, *args, **karg)




from aquests.protocols.http import response

class Response (response.Response):
	def __init__ (self, request, code, msg, opcode, data = None):
		self.request = request
		self.code = code
		self.msg = msg
		self.__data = data
		self.version = "1.1"
		self.opcode = opcode
	
	@property
	def data (self):
		return self.__data
	
	@property
	def headers (self):
		return {}
		
	def done (self):
		pass
	


import odbc, dbi
from . import logger
import sys
import threading
import re

RX_NONASC = re.compile (r"[^\x01-\x80]")
def sqlsafe_value (value):
	global RX_NONASC	
	if value in ('', None):
		value = 'NULL'
	
	elif type (value) == type (''):
		value = RX_NONASC.sub (' ', value)
		value = value.replace (chr(0) ,"").replace ("'", "''").strip ()		
		if not value:
			value = 'NULL'
		else:
			value = "'%s'" % value
			
	else:
		value = str (value)
		
	return value	

def sqlsafe (*values):
	result = []
	for value in values:
		result.append (sqlsafe_value (value))
	return tuple (result)


class NoError (Exception): pass
class IntegrityError (Exception): pass
class ProgramError (Exception): pass
class OperationError (Exception): pass
class InternalError (Exception): pass
class DBServerError (Exception): pass

class QueryLogger:
	def __init__ (self, logger = None):
		self.logger = logger

	def __call__ (self, catch, query):
		self.log (catch, query)

	def log (self, catch, query):
		if self.logger:
			self.logger.log ("%s %s" % (catch, query))

class odbcf:
	def __init__ (self, conninfo, logger = None, log_query = 1):
		self.conninfo = conninfo
		self.logger = logger
		self.log_query = log_query
		self.quqeylogger = QueryLogger (logger)
		self.conn = None
		self.cx = None
		self.cv = threading.Lock ()
		self.connected = 0
		self.autocommit = False

	def __getattr__ (self, attr):
		if self.cx:
			return getattr (self.cx, attr)
		else:
			raise AttributeError
	
	def setautocommit (self, flag):
		self.autocommit = flag		
		
	def log (self, line, type="info", name=""):
		if self.logger:
			self.logger.log (line, type, name)
		else:
			print(line)

	def trace (self, name = ""):
		if self.logger:
			self.logger.trace (name)
		else:
			print(logger.trace ())

	#----------------------------------------------------------------
	# database connect
	#----------------------------------------------------------------
	def connect (self):
		connopt = self.conninfo.split ("/")
		
		try:
			self.conn = odbc.odbc ("/".join (connopt [:3]))
			self.conn.setautocommit (self.autocommit)
			self.cx = self.conn.cursor ()
			self.connected = 1
			
			if len (connopt) == 4:
				self.execute ("use %s" % connopt [-1])
				self.commit ()
				
			self.execute ("set ANSI_NULLS ON; SET ANSI_WARNINGS ON;")
			self.commit ()
			
		except (dbi.opError, dbi.internalError) as why:
			raise DBServerError(str (sys.exc_info()[0]) + ", " + str (why))
		
		except (dbi.noError, dbi.integrityError, dbi.progError) as why:
			self.close ()
			raise DBServerError(str (sys.exc_info()[0]) + ", " + str (why))
		
	def reconnect (self):
		self.close ()
		self.connect ()
	
	def abort (self):
		self.cx = None
		self.conn = None
		self.connected = 0		
		return
		
	def close (self):		
		if not self.connected: 
			return
		
		try:	
			if self.cx:
				self.execute ("set ANSI_NULLS OFF; SET ANSI_WARNINGS OFF;")
				self.commit ()
				try: 
					self.cx.close ()
				except: 
					pass				
	
			if self.conn:
				try: 
					self.conn.close ()					
				except: 
					pass				
	
		finally:
			self.abort ()
			
	def commit (self):
		if not self.autocommit:
			self.conn.commit ()		

	def rollback (self):
		if not self.autocommit:
			self.conn.rollback ()
	
	def handle_expt (self, why, query):	
		errstr = str (why)
		if self.log_query: self.quqeylogger (errstr, query)			
		return errstr
		
	def execute (self, query):
		if not self.connected: 
			raise DBServerError("Not Connected")
			
		try:
			self.cx.execute (query)
		
		except dbi.opError as why:
			errmsg = self.handle_expt (why, query)
			raise OperationError(errmsg)
		
		except dbi.internalError as why:
			errmsg = self.handle_expt (why, query)
			raise InternalError(errmsg)
			
		except dbi.noError as why:
			errmsg = self.handle_expt (why, query)
			raise NoError(errmsg)
			
		except dbi.integrityError as why:
			errmsg = self.handle_expt (why, query)
			raise IntegrityError(errmsg)
		
		except dbi.progError as why:
			errmsg = self.handle_expt (why, query)
			raise ProgramError(errmsg)




from . import response, request
from aquests.protocols.http import request_handler
from aquests.protocols.http.response import FailedResponse
from aquests.client import asynconnect
from aquests.lib import strutil

try:
	from urllib.parse import urlparse
except ImportError:
	from urlparse import urlparse
from base64 import b64encode
import os
import struct
import sys
try:
	from cStringIO import StringIO as BytesIO
except ImportError:
	from io import BytesIO
from . import OPCODE_CLOSE, OPCODE_TEXT, PAYLOAD_LEN, FIN, OPCODE, MASKED
	
class RequestHandler (request_handler.RequestHandler):
	def __init__ (self, asyncon, request, callback, *args, **karg):
		request_handler.RequestHandler.__init__ (self, asyncon, request, callback)
		self.initialize ()
	
	def initialize (self):
		self.buf = b""
		self.rfile = BytesIO ()		
		self.opcode = None
		self.payload_length = 0		
		self.has_masks = True
		self.masks = b""
		self._handshaking = False
		
	def handle_request (self):
		#print ("self.asyncon.upgraded", self.asyncon.upgraded)
		if not self.asyncon.upgraded:
			self.buffer, self.response = b"", None
			self._handshaking = True			
			self.asyncon.set_terminator (b"\r\n\r\n")
			for buf in self.get_request_buffer ():
				self.asyncon.push (buf)
		else:	
			self.asyncon.set_terminator (2)
			self.asyncon.push (self.request.get_message ())
		self.asyncon.begin_tran (self)
		
	def get_request_buffer (self):
		hc = {}		
		scheme, netloc, script, param, queystring, fragment = urlparse (self.request.uri)
		
		addr, port = self.request.get_address ()
		if (scheme == "ws" and port == 80) or (scheme == "wss" and port == 443):
			host = addr [0]
		else:
			host = "%s:%d" % (addr, port)
					
		hc ['Host'] = host
		hc ['Origin'] = "%s://%s" % (type (self.asyncon) is asynconnect.AsynConnect and "https" or "http", hc ['Host'])
		hc ['Sec-WebSocket-Key'] = b64encode(os.urandom(16))
		hc ['Connection'] = "keep-alive, Upgrade"
		hc ['Upgrade'] = 'websocket'
		hc ['Cache-Control'] = 'no-cache'		
		
		auth_header = self.get_http_auth_header ()
		if auth_header:
			hc ["Authorization"] = auth_header
		
		uri = self.asyncon.is_proxy () and self.request.uri.replace ("wss://", "https://").replace ("ws://", "http://") or self.request.path
		req = ("GET %s HTTP/1.1\r\n%s\r\n\r\n" % (
			uri,
			"\r\n".join (["%s: %s" % x for x in list(hc.items ())])
		)).encode ("utf8")		
		return [req]
	
	def connection_closed (self, why, msg):
		if self._handshaking:
			# possibly retry or close_case with error
			request_handler.RequestHandler.connection_closed (self, why, msg)
		else:
			self.response = response.Response (self.request, why, msg, -1, -1)
			self.close_case_with_end_tran ()
	
	def collect_incoming_data (self, data):
		#print ("+++++++", data, self._handshaking)
		if self._handshaking:
			request_handler.RequestHandler.collect_incoming_data (self, data)
		elif self.masks or (not self.has_masks and self.payload_length):
			self.rfile.write (data)
		else:
			self.buf = data	
		
	def _tobytes (self, b):
		if sys.version_info[0] < 3:
			return map(ord, b)
		else:
			return b
	
	def found_end_of_body (self):
		if self.handled_http_authorization ():
			return
		
		if not (self.response.code == 101 and self.response.get_header ("Sec-WebSocket-Accept")):
			self.asyncon.handle_close (self.response.code, self.response.msg)
			
		else:
			self.response = None
			self._handshaking = False
			self.asyncon.upgraded = True											

			self.asyncon.push (self.request.get_message ())
			self.asyncon.set_terminator (2)
						
	def found_terminator (self):
		if self._handshaking:
			request_handler.RequestHandler.found_terminator (self)
			
		elif self.masks or not self.has_masks:
			# end of message
			masked_data = bytearray(self.rfile.getvalue ())
			if self.masks:
				masking_key = bytearray(self.masks)
				data = bytearray ([masked_data[i] ^ masking_key [i%4] for i in range (len (masked_data))])
			else:
				data = masked_data	
			
			if self.opcode == OPCODE_TEXT:
				# text
				data = data.decode('utf-8')
			
			self.response = response.Response (self.request, 200, "OK", self.opcode, data)
			self.asyncon.set_terminator (2)			
			self.close_case_with_end_tran ()
						
		elif self.payload_length:
			self.masks = self.buf
			self.asyncon.set_terminator (self.payload_length)
		
		elif self.opcode:
			if len (self.buf) == 2:
				fmt = ">H"
			else:
				fmt = ">Q"
			self.payload_length = struct.unpack(fmt, self._tobytes(self.buf))[0]
			if self.has_masks:
				self.asyncon.set_terminator (4) # mask
			else:
				self.asyncon.set_terminator (self.payload_length)
		
		elif self.opcode is None:
			b1, b2 = self._tobytes(self.buf)
			fin    = b1 & FIN
			self.opcode = b1 & OPCODE
			if self.opcode == OPCODE_CLOSE:				
				self.response = response.Response (self.request, 200, "OK", self.opcode, "")
				self.asyncon.disconnect ()
				self.close_case_with_end_tran ()
				return
				
			mask = b2 & MASKED
			if not mask:
				self.has_masks = False
			
			payload_length = b2 & PAYLOAD_LEN
			if payload_length == 0:
				self.response = response.Response (self.request, 200, "OK", self.opcode, "")
				self.opcode = None
				self.has_masks = True
				self.asyncon.set_terminator (2)				
				self.close_case_with_end_tran ()
				return
			
			if payload_length < 126:
				self.payload_length = payload_length
				if self.has_masks:
					self.asyncon.set_terminator (4) # mask
				else:
					self.asyncon.set_terminator (self.payload_length)
			elif payload_length == 126:
				self.asyncon.set_terminator (2)	# short length
			elif payload_length == 127:
				self.asyncon.set_terminator (8) # long length

		else:
			raise AssertionError ("Web socket frame decode error")
					
		
